"""Contains the `AxisymmetricWavelets` class."""
import logging

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pys2let

import sleplet._string_methods
import sleplet._validation
import sleplet.wavelet_methods
from sleplet.functions.flm import Flm

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class AxisymmetricWavelets(Flm):
    """
    Create scale-discretised axisymmetric wavelets. As
    seen in <https://doi.org/10.1051/0004-6361/201220729>.
    """

    B: int = 3
    r"""The wavelet parameter. Represented as \(\lambda\) in the papers."""
    j_min: int = 2
    r"""The minimum wavelet scale. Represented as \(J_{0}\) in the papers."""
    j: int | None = None
    """Option to select a given wavelet. `None` indicates the scaling function,
    whereas `0` would correspond to the selected `j_min`."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        _logger.info("start computing wavelets")
        self.wavelets = self._create_wavelets()
        _logger.info("finish computing wavelets")
        jth = 0 if self.j is None else self.j + 1
        return self.wavelets[jth]

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{sleplet._string_methods.filename_args(self.B, 'B')}"
            f"{sleplet._string_methods.filename_args(self.j_min, 'jmin')}"
            f"{sleplet._string_methods.wavelet_ending(self.j_min, self.j)}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return True

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 3
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.B, self.j_min, self.j = self.extra_args

    def _create_wavelets(self: typing_extensions.Self) -> npt.NDArray[np.complex_]:
        """Compute all wavelets."""
        return sleplet.wavelet_methods._create_axisymmetric_wavelets(
            self.L,
            self.B,
            self.j_min,
        )

    @pydantic.field_validator("j")
    def _check_j(
        cls,  # noqa: ANN101
        v: int | None,
        info: pydantic.ValidationInfo,
    ) -> int | None:
        j_max = pys2let.pys2let_j_max(
            info.data["B"],
            info.data["L"],
            info.data["j_min"],
        )
        if v is not None and v < 0:
            msg = "j should be positive"
            raise ValueError(msg)
        if v is not None and v > j_max - info.data["j_min"]:
            msg = (
                "j should be less than j_max - j_min: "
                f"{j_max - info.data['j_min'] + 1}"
            )
            raise ValueError(msg)
        return v



"""Contains the abstract `SlepianFunctions` class."""
import abc
import dataclasses
import logging

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._validation
from sleplet.slepian.region import Region

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class SlepianFunctions:
    """Abstract parent class of creating the different Slepian regions on the sphere."""

    L: int
    """The spherical harmonic bandlimit."""
    eigenvalues: npt.NDArray[np.float_] = dataclasses.field(
        default_factory=lambda: np.empty(0),
        kw_only=True,
        repr=True,
    )
    eigenvectors: npt.NDArray[np.complex_] = dataclasses.field(
        default_factory=lambda: np.empty(0, dtype=np.complex_),
        kw_only=True,
        repr=True,
    )
    mask: npt.NDArray[np.float_] = dataclasses.field(
        default_factory=lambda: np.empty(0),
        kw_only=True,
        repr=False,
    )
    matrix_location: str = dataclasses.field(default="", kw_only=True, repr=False)
    N: int = dataclasses.field(default=0, kw_only=True, repr=False)
    name: str = dataclasses.field(default="", kw_only=True, repr=False)
    region: Region = dataclasses.field(
        default_factory=lambda: Region(theta_max=0),
        kw_only=True,
        repr=False,
    )
    _resolution: int = dataclasses.field(
        default=0,
        kw_only=True,
        repr=False,
    )

    def __post_init__(self: typing_extensions.Self) -> None:
        self.region = self._create_region()
        self.mask = self._create_mask()
        self.name = self._create_fn_name()
        area = self._calculate_area()
        self.N = round(area * self.L**2 / (4 * np.pi))
        msg = f"Shannon number N={self.N}"
        _logger.info(msg)
        self.matrix_location = self._create_matrix_location()
        _logger.info("start solving eigenproblem")
        self.eigenvalues, self.eigenvectors = self._solve_eigenproblem()
        _logger.info("finished solving eigenproblem")

    @abc.abstractmethod
    def _create_fn_name(self: typing_extensions.Self) -> str:
        """Create the name for plotting."""
        raise NotImplementedError

    @abc.abstractmethod
    def _create_region(self: typing_extensions.Self) -> "sleplet.slepian.region.Region":
        """Create a region object for area of interest."""
        raise NotImplementedError

    @abc.abstractmethod
    def _create_mask(self: typing_extensions.Self) -> npt.NDArray[np.float_]:
        """Create a mask of the region of interest."""
        raise NotImplementedError

    @abc.abstractmethod
    def _calculate_area(self: typing_extensions.Self) -> float:
        """Calculate area of region."""
        raise NotImplementedError

    @abc.abstractmethod
    def _create_matrix_location(self: typing_extensions.Self) -> str:
        """Create the name of the matrix binary."""
        raise NotImplementedError

    @abc.abstractmethod
    def _solve_eigenproblem(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """Solve the eigenproblem for the given function."""
        raise NotImplementedError



"""Classes to create and handle mesh (manifold) data."""

from .mesh import Mesh
from .mesh_basis_functions import MeshBasisFunctions
from .mesh_field import MeshField
from .mesh_noise_field import MeshNoiseField
from .mesh_slepian import MeshSlepian
from .mesh_slepian_field import MeshSlepianField
from .mesh_slepian_functions import MeshSlepianFunctions
from .mesh_slepian_noise_field import MeshSlepianNoiseField
from .mesh_slepian_wavelet_coefficients import MeshSlepianWaveletCoefficients
from .mesh_slepian_wavelets import MeshSlepianWavelets

__all__ = [
    "Mesh",
    "MeshBasisFunctions",
    "MeshField",
    "MeshNoiseField",
    "MeshSlepian",
    "MeshSlepianField",
    "MeshSlepianFunctions",
    "MeshSlepianNoiseField",
    "MeshSlepianWaveletCoefficients",
    "MeshSlepianWavelets",
]



"""Contains the abstract `MeshSlepianCoefficients` class."""
import abc

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._validation
import sleplet.noise
from sleplet.meshes.mesh import Mesh
from sleplet.meshes.mesh_coefficients import MeshCoefficients
from sleplet.meshes.mesh_slepian import MeshSlepian


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class MeshSlepianCoefficients(MeshCoefficients):
    """Abstract parent class to handle Slepian coefficients on the mesh."""

    mesh_slepian: MeshSlepian = pydantic.Field(
        default=MeshSlepian(Mesh("bird")),
        init_var=False,
        repr=False,
    )

    def __post_init__(self: typing_extensions.Self) -> None:
        self.mesh_slepian = sleplet.meshes.mesh_slepian.MeshSlepian(self.mesh)
        super().__post_init__()

    def _add_noise_to_signal(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.complex_ | np.float_] | None, float | None]:
        """Add Gaussian white noise converted to Slepian space."""
        self.coefficients: npt.NDArray[np.complex_ | np.float_]
        if self.noise is not None:
            unnoised_coefficients = self.coefficients.copy()
            n_p = sleplet.noise._create_slepian_mesh_noise(
                self.mesh_slepian,
                self.coefficients,
                self.noise,
            )
            snr = sleplet.noise.compute_snr(self.coefficients, n_p, "Slepian")
            self.coefficients = self.coefficients + n_p
            return unnoised_coefficients, snr
        return None, None

    @abc.abstractmethod
    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        raise NotImplementedError

    @abc.abstractmethod
    def _create_name(self: typing_extensions.Self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def _setup_args(self: typing_extensions.Self) -> None:
        raise NotImplementedError



"""Contains the `SlepianDiracDelta` class."""
import logging

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._string_methods
import sleplet._validation
import sleplet._vars
import sleplet.slepian_methods
from sleplet.functions.fp import Fp

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class SlepianDiracDelta(Fp):
    """Create a Dirac delta of the Slepian coefficients."""

    _alpha: float = pydantic.Field(default=0, init_var=False, repr=False)
    _beta: float = pydantic.Field(default=0, init_var=False, repr=False)

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        self._compute_angles()
        return sleplet.slepian_methods._compute_s_p_omega_prime(
            self.L,
            self._alpha,
            self._beta,
            self.slepian,
        ).conj()

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"_{self.slepian.region._name_ending}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return False

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            msg = f"{self.__class__.__name__} does not support extra arguments"
            raise TypeError(msg)

    def _compute_angles(self: typing_extensions.Self) -> None:
        """Compute alpha/beta if not provided."""
        thetas, phis = ssht.sample_positions(
            self.L,
            Grid=True,
            Method=sleplet._vars.SAMPLING_SCHEME,
        )
        sp = ssht.inverse(
            self.slepian.eigenvectors[0],
            self.L,
            Method=sleplet._vars.SAMPLING_SCHEME,
        )
        idx = tuple(np.argwhere(sp == sp.max())[0])
        self._alpha = phis[idx]
        self._beta = thetas[idx]
        msg = (
            f"angles: (alpha, beta) = ({self._alpha/np.pi:.5f},"
            f"{self._beta/np.pi:.5f})\n"
            f"grid point: (alpha, beta) = ({self._alpha:e},{self._beta:e})"
        )
        _logger.info(msg)



"""Contains the `SlepianNoiseAfrica` class."""
import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._string_methods
import sleplet._validation
import sleplet.functions.slepian_africa
import sleplet.noise
import sleplet.slepian.region
from sleplet.functions.fp import Fp


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class SlepianNoiseAfrica(Fp):
    """
    Create a noised Slepian region on the topographic map of the Earth of
    the Africa region.
    """

    SNR: float = -10
    """A parameter which controls the level of signal-to-noise in the noised
    data."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()
        if (
            isinstance(self.region, sleplet.slepian.region.Region)
            and self.region._name_ending != "africa"
        ):
            msg = "Slepian region selected must be 'africa'"
            raise RuntimeError(msg)

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        sa = sleplet.functions.slepian_africa.SlepianAfrica(
            self.L,
            region=self.region,
            smoothing=self.smoothing,
        )
        noise = sleplet.noise._create_slepian_noise(
            self.L,
            sa.coefficients,
            self.slepian,
            self.SNR,
        )
        sleplet.noise.compute_snr(sa.coefficients, noise, "Slepian")
        return noise

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{sleplet._string_methods.filename_args(self.SNR, 'snr')}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return False

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 1
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.SNR = self.extra_args[0]



import dataclasses
import logging

import cmocean
import numpy as np
import numpy.typing as npt
import plotly.graph_objs as go
import plotly.io as pio
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._plotly_methods
import sleplet._validation
import sleplet._vars
import sleplet.plot_methods
import sleplet.slepian.region

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class PlotSphere:
    """Create surface sphere plot via `plotly`."""

    f: npt.NDArray[np.complex_ | np.float_]
    """The field value sampled on the sphere."""
    L: int
    """The spherical harmonic bandlimit."""
    filename: str
    """The output filename of the plot."""
    _: dataclasses.KW_ONLY
    amplitude: float | None = None
    """Whether to customise the amplitude range of the colour bar."""
    annotations: list[dict] = pydantic.Field(default_factory=list)
    """Whether to display any annotations on the surface plot or not."""
    normalise: bool = True
    """Whether to normalise the plot or not."""
    plot_type: str = "real"
    """Whether to display the `real`, `imag`, `abs` or `sum` value of the field."""
    reality: bool = False
    """Whether the given signal is real or not."""
    region: sleplet.slepian.region.Region | None = None
    """Whether to set the field values outside of a given region to zero."""
    spin: int = 0
    """Spin value."""
    upsample: bool = True
    """Whether to upsample the current field."""
    _resolution: int = pydantic.Field(default=0, init_var=False, repr=False)

    def __post_init__(self: typing_extensions.Self) -> None:
        self._resolution = (
            sleplet.plot_methods.calc_plot_resolution(self.L)
            if self.upsample
            else self.L
        )
        if self.upsample:
            self.filename += f"_res{self._resolution}"
        self.filename += f"_{self.plot_type}"
        if self.normalise:
            self.filename += "_norm"

    def execute(self: typing_extensions.Self) -> None:
        """Perform the plot."""
        f = self._prepare_field(self.f)

        # get values from the setup
        x, y, z, f_plot, vmin, vmax = self._setup_plot(
            f,
            self._resolution,
            method=sleplet._vars.SAMPLING_SCHEME,
        )

        if isinstance(self.region, sleplet.slepian.region.Region):
            # make plot area clearer
            f_plot = sleplet.plot_methods._set_outside_region_to_minimum(
                f_plot,
                self._resolution,
                self.region,
            )

        # appropriate zoom in on north pole
        camera = sleplet._plotly_methods.create_camera(-0.1, -0.1, 10, 7.88)

        # pick largest tick max value
        tick_mark = sleplet._plotly_methods.create_tick_mark(
            vmin,
            vmax,
            amplitude=self.amplitude,
        )

        data = [
            go.Surface(
                x=x,
                y=y,
                z=z,
                surfacecolor=f_plot,
                cmax=1 if self.normalise else tick_mark,
                cmid=0.5 if self.normalise else 0,
                cmin=0 if self.normalise else -tick_mark,
                colorbar=sleplet._plotly_methods.create_colour_bar(
                    tick_mark,
                    normalise=self.normalise,
                ),
                colorscale=sleplet.plot_methods._convert_colourscale(cmocean.cm.ice),
                lighting=go.surface.Lighting(ambient=1),
                reversescale=True,
            ),
        ]

        layout = sleplet._plotly_methods.create_layout(
            camera,
            annotations=self.annotations,
        )

        fig = go.Figure(data=data, layout=layout)

        msg = f"Opening: {self.filename}"
        _logger.info(msg)
        pio.show(fig, config={"toImageButtonOptions": {"filename": self.filename}})

    @staticmethod
    def _setup_plot(  # noqa: PLR0913
        f: npt.NDArray[np.float_],
        resolution: int,
        *,
        method: str = "MW",
        close: bool = True,
        parametric: bool = False,
        parametric_scaling: list[float] | None = None,
        color_range: list[float] | None = None,
    ) -> tuple[
        npt.NDArray[np.float_],
        npt.NDArray[np.float_],
        npt.NDArray[np.float_],
        npt.NDArray[np.float_],
        float,
        float,
    ]:
        """Create the data for the matplotlib/plotly plot."""
        if parametric_scaling is None:
            parametric_scaling = [0.0, 0.5]

        thetas, phis = ssht.sample_positions(resolution, Grid=True, Method=method)

        if thetas.size != f.size:
            msg = "Bandlimit L deos not match that of f"
            raise AttributeError(msg)

        f_plot = f.copy()

        f_max = f_plot.max()
        f_min = f_plot.min()

        if not isinstance(color_range, list):
            vmin = f_min
            vmax = f_max
        else:
            vmin = color_range[0]
            vmax = color_range[1]
            f_plot[f_plot < color_range[0]] = color_range[0]
            f_plot[f_plot > color_range[1]] = color_range[1]
            f_plot[f_plot == sleplet._vars.SPHERE_UNSEEN] = np.nan

        # Compute position scaling for parametric plot.
        f_normalised = (
            (f_plot - vmin / (vmax - vmin)) * parametric_scaling[1]
            + parametric_scaling[0]
            if parametric
            else np.zeros(f_plot.shape)
        )

        # Close plot.
        if close:
            _, n_phi = ssht.sample_shape(resolution, Method=method)
            f_plot = np.insert(f_plot, n_phi, f[:, 0], axis=1)
            if parametric:
                f_normalised = np.insert(
                    f_normalised,
                    n_phi,
                    f_normalised[:, 0],
                    axis=1,
                )
            thetas = np.insert(thetas, n_phi, thetas[:, 0], axis=1)
            phis = np.insert(phis, n_phi, phis[:, 0], axis=1)

        # Compute location of vertices.
        x, y, z = (
            ssht.spherical_to_cart(f_normalised, thetas, phis)
            if parametric
            else ssht.s2_to_cart(thetas, phis)
        )

        return x, y, z, f_plot, vmin, vmax

    def _prepare_field(
        self: typing_extensions.Self,
        f: npt.NDArray[np.complex_ | np.float_],
    ) -> npt.NDArray[np.float_]:
        """Boosts, forces plot type and then scales the field before plotting."""
        boosted_field = sleplet.plot_methods._boost_field(
            f,
            self.L,
            self._resolution,
            reality=self.reality,
            spin=self.spin,
            upsample=self.upsample,
        )
        field_space = sleplet.plot_methods._create_plot_type(
            boosted_field,
            self.plot_type,
        )
        return sleplet.plot_methods._normalise_function(
            field_space,
            normalise=self.normalise,
        )



"""Contains the `MeshSlepian` class."""
import concurrent.futures
import logging
import os

import numpy as np
import numpy.linalg as LA  # noqa: N812
import numpy.typing as npt
import platformdirs
import pydantic
import typing_extensions

import sleplet._array_methods
import sleplet._data.setup_pooch
import sleplet._integration_methods
import sleplet._parallel_methods
import sleplet._slepian_arbitrary_methods
import sleplet._validation
from sleplet.meshes.mesh import Mesh

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class MeshSlepian:
    """Create Slepian object of a given mesh."""

    mesh: Mesh
    """A mesh object."""
    N: int = pydantic.Field(default=0, init_var=False, repr=False)
    slepian_eigenvalues: npt.NDArray[np.float_] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )
    slepian_functions: npt.NDArray[np.float_] = pydantic.Field(
        default_factory=lambda: np.empty(0),
        init_var=False,
        repr=False,
    )

    def __post_init__(self: typing_extensions.Self) -> None:
        self.N = sleplet._slepian_arbitrary_methods.compute_mesh_shannon(self.mesh)
        self._compute_slepian_functions()

    def _compute_slepian_functions(self: typing_extensions.Self) -> None:
        """Compute the Slepian functions of the mesh."""
        _logger.info("computing slepian functions of mesh")

        # create filenames
        eigd_loc = (
            f"meshes_laplacians_slepian_functions_{self.mesh.name}_"
            f"b{self.mesh.mesh_eigenvalues.shape[0]}_N{self.N}"
        )
        eval_loc = f"{eigd_loc}_eigenvalues.npy"
        evec_loc = f"{eigd_loc}_eigenvectors.npy"

        try:
            self.slepian_eigenvalues = np.load(
                sleplet._data.setup_pooch.find_on_pooch_then_local(eval_loc),
            )
            self.slepian_functions = np.load(
                sleplet._data.setup_pooch.find_on_pooch_then_local(evec_loc),
            )
        except TypeError:
            self._compute_slepian_functions_from_scratch(eval_loc, evec_loc)

    def _compute_slepian_functions_from_scratch(
        self: typing_extensions.Self,
        eval_loc: str,
        evec_loc: str,
    ) -> None:
        D = self._create_D_matrix()
        msg = (
            f"Shannon number from vertices: {self.N}, "
            f"Trace of D matrix: {round(D.trace())}, "
            f"difference: {round(np.abs(self.N - D.trace()))}",
        )
        _logger.info(msg)

        # fill in remaining triangle section
        sleplet._array_methods.fill_upper_triangle_of_hermitian_matrix(D)

        # solve eigenproblem
        (
            self.slepian_eigenvalues,
            self.slepian_functions,
        ) = self._clean_evals_and_evecs(LA.eigh(D))
        np.save(platformdirs.user_data_path() / eval_loc, self.slepian_eigenvalues)
        np.save(
            platformdirs.user_data_path() / evec_loc,
            self.slepian_functions[: self.N],
        )

    def _create_D_matrix(  # noqa: N802
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.float_]:
        """Compute the D matrix for the mesh eigenfunctions."""
        D = np.zeros(
            (self.mesh.mesh_eigenvalues.shape[0], self.mesh.mesh_eigenvalues.shape[0]),
        )

        D_ext, shm_ext = sleplet._parallel_methods.create_shared_memory_array(D)

        def func(chunk: list[int]) -> None:
            """Calculate D matrix components for each chunk."""
            D_int, shm_int = sleplet._parallel_methods.attach_to_shared_memory_block(
                D,
                shm_ext,
            )

            for i in chunk:
                msg = f"start basis function: {i}"
                _logger.info(msg)
                self._fill_D_elements(D_int, i)
                msg = f"finish basis function: {i}"
                _logger.info(msg)

            sleplet._parallel_methods.free_shared_memory(shm_int)

        # split up L range to maximise efficiency
        ncpu = int(os.getenv("NCPU", "4"))
        msg = f"Number of CPU={ncpu}"
        _logger.info(msg)
        chunks = sleplet._parallel_methods.split_arr_into_chunks(
            self.mesh.mesh_eigenvalues.shape[0],
            ncpu,
        )

        # initialise pool and apply function
        with concurrent.futures.ThreadPoolExecutor(max_workers=ncpu) as e:
            e.map(func, chunks)

        # retrieve from parallel function
        D = D_ext.copy()

        # Free and release the shared memory block at the very end
        sleplet._parallel_methods.free_shared_memory(shm_ext)
        sleplet._parallel_methods.release_shared_memory(shm_ext)
        return D

    def _fill_D_elements(  # noqa: N802
        self: typing_extensions.Self,
        D: npt.NDArray[np.float_],
        i: int,
    ) -> None:
        """Fill in the D matrix elements using symmetries."""
        D[i][i] = self._integral(i, i)
        for j in range(i + 1, self.mesh.mesh_eigenvalues.shape[0]):
            D[j][i] = self._integral(j, i)

    def _integral(self: typing_extensions.Self, i: int, j: int) -> float:
        """Calculate the D integral between two mesh basis functions."""
        return sleplet._integration_methods.integrate_region_mesh(
            self.mesh.mesh_region,
            self.mesh.vertices,
            self.mesh.faces,
            self.mesh.basis_functions[i],
            self.mesh.basis_functions[j],
        )

    @staticmethod
    def _clean_evals_and_evecs(
        eigendecomposition: tuple,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        """Need eigenvalues and eigenvectors to be in a certain format."""
        # access values
        eigenvalues, eigenvectors = eigendecomposition

        # sort eigenvalues and eigenvectors in descending order of eigenvalues
        idx = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[idx]
        eigenvectors = eigenvectors[:, idx].T
        return eigenvalues, eigenvectors



""".. include:: ./../../documentation/DOCUMENTATION.md"""  # noqa: D400,D415

import logging

from . import (
    functions,
    harmonic_methods,
    meshes,
    noise,
    plot_methods,
    plotting,
    slepian,
    slepian_methods,
    wavelet_methods,
)
from ._version import __version__  # noqa: F401

__all__ = [
    "functions",
    "harmonic_methods",
    "meshes",
    "noise",
    "plot_methods",
    "plotting",
    "slepian_methods",
    "slepian",
    "wavelet_methods",
]

_logger = logging.getLogger(__name__)

_formatter = logging.Formatter(
    "%(levelname)s [%(asctime)s] sleplet: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
_console_handler = logging.StreamHandler()
_console_handler.setFormatter(_formatter)
_logger.addHandler(_console_handler)
_logger.setLevel("INFO")
_logger.propagate = False



"""Contains the `MeshSlepianFunctions` class."""
import logging

import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import sleplet._validation
import sleplet.slepian_methods
from sleplet.meshes.mesh_slepian_coefficients import MeshSlepianCoefficients

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class MeshSlepianFunctions(MeshSlepianCoefficients):
    """Create Slepian functions of a given mesh."""

    rank: int = 0
    """Slepian eigenvalues are ordered in decreasing value. The option `rank`
    selects a given Slepian function from the spectrum (p in the papers)."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        """Compute field on the vertices of the mesh."""
        msg = (
            f"Slepian eigenvalue {self.rank}: "
            f"{self.mesh_slepian.slepian_eigenvalues[self.rank]:e}",
        )
        _logger.info(msg)
        s_p_i = self.mesh_slepian.slepian_functions[self.rank]
        return sleplet.slepian_methods.slepian_mesh_forward(
            self.mesh_slepian,
            u_i=s_p_i,
        )

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            (
                f"slepian_{self.mesh.name}_rank{self.rank}_"
                f"lam{self.mesh_slepian.slepian_eigenvalues[self.rank]:e}"
            )
            .replace(".", "-")
            .replace("+", "")
        )

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 1
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be 1 or {num_args}"
                raise ValueError(msg)
            self.rank = self.extra_args[0]

    def _validate_rank(self: typing_extensions.Self) -> None:
        """Check the requested rank is valid."""
        if isinstance(self.extra_args, list):
            limit = self.mesh.mesh_eigenvalues.shape[0]
            if self.extra_args[0] > limit:
                msg = f"rank should be less than or equal to {limit}"
                raise ValueError(msg)

    @pydantic.field_validator("rank")
    def _check_rank(
        cls,  # noqa: ANN101
        v: int,
    ) -> int:
        if not isinstance(v, int):
            msg = "rank should be an integer"
            raise TypeError(msg)
        if v < 0:
            msg = "rank cannot be negative"
            raise ValueError(msg)
        return v



"""Methods to work with wavelet and wavelet coefficients."""
import numpy as np
import numpy.typing as npt

import pys2let
import pyssht as ssht

import sleplet._convolution_methods
import sleplet.slepian_methods


def slepian_wavelet_forward(
    f_p: npt.NDArray[np.complex_ | np.float_],
    wavelets: npt.NDArray[np.float_],
    shannon: int,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    Compute the coefficients of the given tiling function in Slepian space.

    Args:
        f_p: The Slepian coefficients.
        wavelets: The Slepian wavelets.
        shannon: The Shannon number.

    Returns:
        The Slepian wavelets coefficients of the signal.
    """
    return find_non_zero_wavelet_coefficients(
        sleplet._convolution_methods.sifting_convolution(
            wavelets,
            f_p,
            shannon=shannon,
        ),
        axis=1,
    )


def slepian_wavelet_inverse(
    wav_coeffs: npt.NDArray[np.complex_ | np.float_],
    wavelets: npt.NDArray[np.float_],
    shannon: int,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    Compute the inverse wavelet transform in Slepian space.

    Args:
        wav_coeffs: The Slepian wavelet coefficients.
        wavelets: The Slepian wavelets.
        shannon: The Shannon number.

    Returns:
        The coefficients of the signal in Slepian space.
    """
    # ensure wavelets are the same shape as the coefficients
    wavelets_shannon = wavelets[: len(wav_coeffs)]
    wavelet_reconstruction = sleplet._convolution_methods.sifting_convolution(
        wavelets_shannon,
        wav_coeffs,
        shannon=shannon,
    )
    return wavelet_reconstruction.sum(axis=0)


def axisymmetric_wavelet_forward(
    L: int,
    flm: npt.NDArray[np.complex_ | np.float_],
    wavelets: npt.NDArray[np.complex_],
) -> npt.NDArray[np.complex_]:
    """
    Compute the coefficients of the axisymmetric wavelets.

    Args:
        L: The spherical harmonic bandlimit.
        flm: The spherical harmonic coefficients.
        wavelets: Axisymmetric wavelets.

    Returns:
        Axisymmetric wavelets coefficients.
    """
    w = np.zeros(wavelets.shape, dtype=np.complex_)
    for ell in range(L):
        ind_m0 = ssht.elm2ind(ell, 0)
        wav_0 = np.sqrt((4 * np.pi) / (2 * ell + 1)) * wavelets[:, ind_m0].conj()
        for m in range(-ell, ell + 1):
            ind = ssht.elm2ind(ell, m)
            w[:, ind] = wav_0 * flm[ind]
    return w


def axisymmetric_wavelet_inverse(
    L: int,
    wav_coeffs: npt.NDArray[np.complex_],
    wavelets: npt.NDArray[np.complex_],
) -> npt.NDArray[np.complex_]:
    """
    Compute the inverse axisymmetric wavelet transform.

    Args:
        L: The spherical harmonic bandlimit.
        wav_coeffs: Axisymmetric wavelet coefficients.
        wavelets: Axisymmetric wavelets.

    Returns:
        Spherical harmonic coefficients of the signal.
    """
    flm = np.zeros(L**2, dtype=np.complex_)
    for ell in range(L):
        ind_m0 = ssht.elm2ind(ell, 0)
        wav_0 = np.sqrt((4 * np.pi) / (2 * ell + 1)) * wavelets[:, ind_m0]
        for m in range(-ell, ell + 1):
            ind = ssht.elm2ind(ell, m)
            flm[ind] = (wav_coeffs[:, ind] * wav_0).sum()
    return flm


def _create_axisymmetric_wavelets(
    L: int,
    B: int,
    j_min: int,
) -> npt.NDArray[np.complex_]:
    """Compute the axisymmetric wavelets."""
    kappas = create_kappas(L, B, j_min)
    wavelets = np.zeros((kappas.shape[0], L**2), dtype=np.complex_)
    for ell in range(L):
        factor = np.sqrt((2 * ell + 1) / (4 * np.pi))
        ind = ssht.elm2ind(ell, 0)
        wavelets[:, ind] = factor * kappas[:, ell]
    return wavelets


def create_kappas(xlim: int, B: int, j_min: int) -> npt.NDArray[np.float_]:
    r"""
    Compute the Slepian wavelets.

    Args:
        xlim: The x-axis value. \(L\) or \(L^2\) in the harmonic or Slepian case.
        B: The wavelet parameter. Represented as \(\lambda\) in the papers.
        j_min: The minimum wavelet scale. Represented as \(J_{0}\) in the papers.

    Returns:
        The Slepian wavelet generating functions.
    """
    kappa0, kappa = pys2let.axisym_wav_l(B, xlim, j_min)
    return np.concatenate((kappa0[np.newaxis], kappa.T))


def find_non_zero_wavelet_coefficients(
    wav_coeffs: npt.NDArray[np.complex_ | np.float_],
    *,
    axis: int | tuple[int, ...],
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    Find the coefficients within the shannon number to speed up computations.

    Args:
        wav_coeffs: The wavelet coefficients.
        axis: The axis to search over.

    Returns:
        The non-zero wavelet coefficients.
    """
    return wav_coeffs[wav_coeffs.any(axis=axis)]



import logging
import pathlib

import igl
import numpy as np
import numpy.typing as npt
import platformdirs
import scipy.sparse.linalg as LA_sparse  # noqa: N812
import tomli

import sleplet._data.setup_pooch
import sleplet._integration_methods

_data_path = pathlib.Path(__file__).resolve().parent / "_data"
_logger = logging.getLogger(__name__)


def average_functions_on_vertices_to_faces(
    faces: npt.NDArray[np.int_],
    functions_on_vertices: npt.NDArray[np.complex_ | np.float_],
) -> npt.NDArray[np.float_]:
    """
    Require all functions to be defined on faces
    this method handles an arbitrary number of functions.
    """
    _logger.info("converting function on vertices to faces")
    # handle the case of a 1D array
    array_is_1d = len(functions_on_vertices.shape) == 1
    if array_is_1d:
        functions_on_vertices = functions_on_vertices.reshape(1, -1)

    functions_on_faces = np.zeros((functions_on_vertices.shape[0], faces.shape[0]))
    for i, f in enumerate(functions_on_vertices):
        functions_on_faces[i] = igl.average_onto_faces(faces, f)

    # put the vector back in 1D form
    if array_is_1d:
        functions_on_faces = functions_on_faces.reshape(-1)
    return functions_on_faces


def create_mesh_region(
    mesh_config: dict,
    vertices: npt.NDArray[np.float_],
) -> npt.NDArray[np.bool_]:
    """Create a boolean region for the given mesh."""
    return (
        (vertices[:, 0] >= mesh_config["XMIN"])
        & (vertices[:, 0] <= mesh_config["XMAX"])
        & (vertices[:, 1] >= mesh_config["YMIN"])
        & (vertices[:, 1] <= mesh_config["YMAX"])
        & (vertices[:, 2] >= mesh_config["ZMIN"])
        & (vertices[:, 2] <= mesh_config["ZMAX"])
    )


def extract_mesh_config(mesh_name: str) -> dict:
    """Read in the given mesh region settings file."""
    with pathlib.Path.open(
        _data_path / f"meshes_regions_{mesh_name}.toml",
        "rb",
    ) as f:
        return tomli.load(f)


def mesh_eigendecomposition(
    name: str,
    vertices: npt.NDArray[np.float_],
    faces: npt.NDArray[np.int_],
    *,
    number_basis_functions: int | None = None,
) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], int]:
    """
    Compute the eigendecomposition of the mesh represented
    as a graph if already computed then it loads the data.
    """
    # determine number of basis functions
    if number_basis_functions is None:
        number_basis_functions = vertices.shape[0] // 4
    msg = (
        f"finding {number_basis_functions}/{vertices.shape[0]} "
        f"basis functions of {name} mesh",
    )
    _logger.info(msg)

    # create filenames
    eigd_loc = f"meshes_laplacians_basis_functions_{name}_b{number_basis_functions}"
    eval_loc = f"{eigd_loc}_eigenvalues.npy"
    evec_loc = f"{eigd_loc}_eigenvectors.npy"

    try:
        eigenvalues = np.load(
            sleplet._data.setup_pooch.find_on_pooch_then_local(eval_loc),
        )
        eigenvectors = np.load(
            sleplet._data.setup_pooch.find_on_pooch_then_local(evec_loc),
        )
    except TypeError:
        laplacian = _mesh_laplacian(vertices, faces)
        eigenvalues, eigenvectors = LA_sparse.eigsh(
            laplacian,
            k=number_basis_functions,
            which="LM",
            sigma=0,
        )
        eigenvectors = _orthonormalise_basis_functions(vertices, faces, eigenvectors.T)
        _logger.info("saving binaries...")
        np.save(platformdirs.user_data_path() / eval_loc, eigenvalues)
        np.save(platformdirs.user_data_path() / evec_loc, eigenvectors)
    return eigenvalues, eigenvectors, number_basis_functions


def read_mesh(mesh_config: dict) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.int_]]:
    """Read in the given mesh."""
    vertices, faces = igl.read_triangle_mesh(
        str(_data_path / f"meshes_polygons_{mesh_config['FILENAME']}"),
    )
    return igl.upsample(vertices, faces, number_of_subdivs=mesh_config["UPSAMPLE"])


def _mesh_laplacian(
    vertices: npt.NDArray[np.float_],
    faces: npt.NDArray[np.int_],
) -> npt.NDArray[np.float_]:
    """Compute the cotagent mesh laplacian."""
    return -igl.cotmatrix(vertices, faces)


def _orthonormalise_basis_functions(
    vertices: npt.NDArray[np.float_],
    faces: npt.NDArray[np.int_],
    basis_functions: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """For computing the Slepian D matrix the basis functions must be orthonormal."""
    _logger.info("orthonormalising basis functions")
    factor = np.zeros(basis_functions.shape[0])
    for i, phi_i in enumerate(basis_functions):
        factor[i] = sleplet._integration_methods.integrate_whole_mesh(
            vertices,
            faces,
            phi_i,
            phi_i,
        )
    normalisation = np.sqrt(factor).reshape(-1, 1)
    return basis_functions / normalisation



"""Contains the `SlepianPolarCap` class."""
import concurrent.futures
import dataclasses
import logging
import os

import gmpy2 as gp
import numpy as np
import numpy.linalg as LA  # noqa: N812
import numpy.typing as npt
import platformdirs
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._data.setup_pooch
import sleplet._mask_methods
import sleplet._parallel_methods
import sleplet._validation
import sleplet.harmonic_methods
import sleplet.slepian.region
from sleplet.slepian.slepian_functions import SlepianFunctions

_logger = logging.getLogger(__name__)

_L_SAVE_ALL = 16


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation)
class SlepianPolarCap(SlepianFunctions):
    """Class to create a polar cap Slepian region on the sphere."""

    theta_max: float
    """Set the size of the polar cap region."""
    _: dataclasses.KW_ONLY
    gap: bool = False
    """Whether to enable a double ended polar cap."""
    order: int | npt.NDArray[np.int_] | None = None
    """
    By default (i.e. `None`) all orders (i.e. `m`) will be computed. If
    `order` is specified by an integer then only a given order `m` will be
    computed. In the Slepian eigenproblem formulation this simplifies the
    mathematical formulation."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_fn_name(self: typing_extensions.Self) -> str:
        return f"slepian_{self.region._name_ending}"

    def _create_region(self: typing_extensions.Self) -> "sleplet.slepian.region.Region":
        return sleplet.slepian.region.Region(gap=self.gap, theta_max=self.theta_max)

    def _create_mask(self: typing_extensions.Self) -> npt.NDArray[np.float_]:
        return sleplet._mask_methods.create_mask_region(self.L, self.region)

    def _calculate_area(self: typing_extensions.Self) -> float:
        return 2 * np.pi * (1 - np.cos(self.theta_max))

    def _create_matrix_location(self: typing_extensions.Self) -> str:
        return (
            f"slepian_eigensolutions_D_{self.region._name_ending}_L{self.L}_N{self.N}"
        )

    def _solve_eigenproblem(
        self: typing_extensions.Self,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        eval_loc = f"{self.matrix_location}_eigenvalues.npy"
        evec_loc = f"{self.matrix_location}_eigenvectors.npy"
        order_loc = f"{self.matrix_location}_orders.npy"
        try:
            eigenvalues, eigenvectors = self._solve_eigenproblem_from_files(
                eval_loc,
                evec_loc,
                order_loc,
            )
        except TypeError:
            eigenvalues, eigenvectors = self._solve_eigenproblem_from_scratch(
                eval_loc,
                evec_loc,
                order_loc,
            )
        return eigenvalues, eigenvectors

    def _solve_eigenproblem_from_files(
        self: typing_extensions.Self,
        eval_loc: str,
        evec_loc: str,
        order_loc: str,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """Solve eigenproblem with files already saved."""
        eigenvalues = np.load(
            sleplet._data.setup_pooch.find_on_pooch_then_local(eval_loc),
        )
        eigenvectors = np.load(
            sleplet._data.setup_pooch.find_on_pooch_then_local(evec_loc),
        )
        orders = np.load(sleplet._data.setup_pooch.find_on_pooch_then_local(order_loc))

        if self.order is not None:
            idx = np.where(orders == self.order)
            return eigenvalues[idx], eigenvectors[idx]
        self.order = orders
        return eigenvalues, eigenvectors

    def _solve_eigenproblem_from_scratch(
        self: typing_extensions.Self,
        eval_loc: str,
        evec_loc: str,
        order_loc: str,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """Solve eigenproblem from scratch and then saves the files."""
        if isinstance(self.order, int):
            return self._solve_eigenproblem_order(self.order)

        evals_all = np.empty(0)
        evecs_all = np.empty((0, self.L**2), dtype=np.complex_)
        emm = np.empty(0, dtype=int)
        for m in range(-(self.L - 1), self.L):
            evals_m, evecs_m = self._solve_eigenproblem_order(m)
            evals_all = np.append(evals_all, evals_m)
            evecs_all = np.concatenate((evecs_all, evecs_m))
            emm = np.append(emm, [m] * len(evals_m))
        (
            eigenvalues,
            eigenvectors,
            self.order,
        ) = self._sort_all_evals_and_evecs(evals_all, evecs_all, emm)
        limit = self.N if self.L > _L_SAVE_ALL else None
        np.save(platformdirs.user_data_path() / eval_loc, eigenvalues)
        np.save(platformdirs.user_data_path() / evec_loc, eigenvectors[:limit])
        np.save(platformdirs.user_data_path() / order_loc, self.order)
        return eigenvalues, eigenvectors

    def _solve_eigenproblem_order(
        self: typing_extensions.Self,
        m: int,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """Solve the eigenproblem for a given order m."""
        emm = sleplet.harmonic_methods._create_emm_vector(self.L)
        Dm = self._create_Dm_matrix(abs(m), emm)
        eigenvalues, gl = LA.eigh(Dm)
        eigenvalues, eigenvectors = self._clean_evals_and_evecs(eigenvalues, gl, emm, m)
        return eigenvalues, eigenvectors

    def _sort_all_evals_and_evecs(
        self: typing_extensions.Self,
        eigenvalues: npt.NDArray[np.float_],
        eigenvectors: npt.NDArray[np.complex_],
        orders: npt.NDArray[np.int_],
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_], npt.NDArray[np.int_]]:
        """Sort all eigenvalues and eigenvectors for all orders."""
        idx = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[idx]
        eigenvectors = eigenvectors[idx]
        orders = orders[idx]
        return eigenvalues, eigenvectors, orders

    def _create_Dm_matrix(  # noqa: N802
        self: typing_extensions.Self,
        m: int,
        emm: npt.NDArray[np.float_],
    ) -> npt.NDArray[np.float_]:
        """
        Syntax:
        Dm = _create_Dm_matrix(m, P).

        Input:
        m  =  order
        P(:,1)  =  Pl = Legendre Polynomials column vector for l = 0 : L-1
        P(:,2)  =  ell values vector

        Output:
        Dm = (L - m) square Slepian matrix for order m

        Description:
        This piece of code computes the Slepian matrix, Dm, for order m and all
        degrees, using the formulation given in "Spatiospectral Concentration on
        a Sphere" by F.J. Simons, F.A. Dahlen and M.A. Wieczorek.
        """
        Pl, ell = self._create_legendre_polynomials_table(emm)
        Dm = np.zeros((self.L - m, self.L - m))
        lvec = np.arange(m, self.L)

        Dm_ext, shm_ext = sleplet._parallel_methods.create_shared_memory_array(Dm)

        def func(chunk: list[int]) -> None:
            """Calculate D matrix components for each chunk."""
            Dm_int, shm_int = sleplet._parallel_methods.attach_to_shared_memory_block(
                Dm,
                shm_ext,
            )

            # deal with chunk
            for i in chunk:
                msg = f"start ell: {i}"
                _logger.info(msg)
                self._dm_matrix_helper(Dm_int, i, m, lvec, Pl, ell)
                msg = f"finish ell: {i}"
                _logger.info(msg)

            sleplet._parallel_methods.free_shared_memory(shm_int)

        # split up L range to maximise efficiency
        ncpu = int(os.getenv("NCPU", "4"))
        msg = f"Number of CPU={ncpu}"
        _logger.info(msg)
        chunks = sleplet._parallel_methods.split_arr_into_chunks(
            self.L - m,
            ncpu,
        )

        # initialise pool and apply function
        with concurrent.futures.ThreadPoolExecutor(max_workers=ncpu) as e:
            e.map(func, chunks)

        # retrieve from parallel function
        Dm = Dm_ext * (-1) ** m / 2

        # Free and release the shared memory block at the very end
        sleplet._parallel_methods.free_shared_memory(shm_ext)
        sleplet._parallel_methods.release_shared_memory(shm_ext)
        return Dm

    def _create_legendre_polynomials_table(
        self: typing_extensions.Self,
        emm: npt.NDArray[np.float_],
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.int_]]:
        """Create Legendre polynomials table for matrix calculation."""
        Plm = ssht.create_ylm(self.theta_max, 0, 2 * self.L).real.reshape(-1)
        ind = emm == 0
        ell = np.arange(2 * self.L)[np.newaxis]
        Pl = np.sqrt((4 * np.pi) / (2 * ell + 1)) * Plm[ind]
        return Pl, ell

    def _dm_matrix_helper(  # noqa: PLR0913
        self: typing_extensions.Self,
        Dm: npt.NDArray[np.float_],
        i: int,
        m: int,
        lvec: npt.NDArray[np.int_],
        Pl: npt.NDArray[np.float_],
        ell: npt.NDArray[np.int_],
    ) -> None:
        """Use in both serial and parallel calculations."""
        el = int(lvec[i])
        for j in range(i, self.L - m):
            p = int(lvec[j])
            c = 0
            for n in range(abs(el - p), el + p + 1):
                A = Pl[ell == n - 1] if n != 0 else 1
                c += (
                    self._wigner3j(el, n, p, 0, 0, 0)
                    * self._wigner3j(el, n, p, m, 0, -m)
                    * (A - Pl[ell == n + 1])
                )
            Dm[i, j] = (
                self._polar_gap_modification(el, p)
                * np.sqrt((2 * el + 1) * (2 * p + 1))
                * c
            )
            Dm[j, i] = Dm[i, j]

    @staticmethod
    def _wigner3j(  # noqa: PLR0913
        l1: int,
        l2: int,
        l3: int,
        m1: int,
        m2: int,
        m3: int,
    ) -> float:
        """
        Syntax:
        s = _wigner3j (l1, l2, l3, m1, m2, m3).

        Input:
        l1  =  first degree in Wigner 3j symbol
        l2  =  second degree in Wigner 3j symbol
        l3  =  third degree in Wigner 3j symbol
        m1  =  first order in Wigner 3j symbol
        m2  =  second order in Wigner 3j symbol
        m3  =  third order in Wigner 3j symbol

        Output:
        s  =  Wigner 3j symbol for l1,m1; l2,m2; l3,m3

        Description:
        Computes Wigner 3j symbol using Racah formula
        """
        if (
            2 * l1 != np.floor(2 * l1)
            or 2 * l2 != np.floor(2 * l2)
            or 2 * l3 != np.floor(2 * l3)
            or 2 * m1 != np.floor(2 * m1)
            or 2 * m2 != np.floor(2 * m2)
            or 2 * m3 != np.floor(2 * m3)
        ):
            msg = "Arguments must either be integer or half-integer!"
            raise ValueError(msg)

        if (
            m1 + m2 + m3 != 0
            or l3 < abs(l1 - l2)
            or l3 > (l1 + l2)
            or abs(m1) > abs(l1)
            or abs(m2) > abs(l2)
            or abs(m3) > abs(l3)
            or l1 + l2 + l3 != np.floor(l1 + l2 + l3)
        ):
            s = 0
        else:
            t1 = l2 - l3 - m1
            t2 = l1 - l3 + m2
            t3 = l1 + l2 - l3
            t4 = l1 - m1
            t5 = l2 + m2

            tmin = max(0, t1, t2)
            tmax = min(t3, t4, t5)

            # sum is over all those t for which the
            # following factorials have non-zero arguments
            s = sum(
                (-1) ** t
                / (
                    gp.factorial(t)
                    * gp.factorial(t - t1)
                    * gp.factorial(t - t2)
                    * gp.factorial(t3 - t)
                    * gp.factorial(t4 - t)
                    * gp.factorial(t5 - t)
                )
                for t in range(tmin, tmax + 1)
            )
            triangle_coefficient = (
                gp.factorial(t3)
                * gp.factorial(l1 - l2 + l3)
                * gp.factorial(-l1 + l2 + l3)
                / gp.factorial(l1 + l2 + l3 + 1)
            )

            s *= (
                np.float_power(-1, l1 - l2 - m3)
                * gp.sqrt(triangle_coefficient)
                * gp.sqrt(
                    gp.factorial(l1 + m1)
                    * gp.factorial(t4)
                    * gp.factorial(t5)
                    * gp.factorial(l2 - m2)
                    * gp.factorial(l3 + m3)
                    * gp.factorial(l3 - m3),
                )
            )
        return s

    def _polar_gap_modification(
        self: typing_extensions.Self,
        ell1: int,
        ell2: int,
    ) -> int:
        """
        Eq 67 - Spherical Slepian functions and the polar gap in geodesy
        multiply by 1 + (-1)*(ell+ell').
        """
        return 1 + self.gap * (-1) ** (ell1 + ell2)

    def _clean_evals_and_evecs(
        self: typing_extensions.Self,
        eigenvalues: npt.NDArray[np.float_],
        gl: npt.NDArray[np.float_],
        emm: npt.NDArray[np.float_],
        m: int,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """Need eigenvalues and eigenvectors to be in a certain format."""
        # Sort eigenvalues and eigenvectors in descending order of eigenvalues
        idx = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[idx]
        gl = gl[:, idx].conj()

        # put back in full D space for harmonic transform
        emm = emm[: self.L**2]
        ind = np.tile(emm == m, (self.L - abs(m), 1))
        eigenvectors = np.zeros((self.L - abs(m), self.L**2), dtype=np.complex_)
        eigenvectors[ind] = gl.T.flatten()

        # ensure first element of each eigenvector is positive
        eigenvectors *= np.where(eigenvectors[:, 0] < 0, -1, 1)[:, np.newaxis]

        # if -ve 'm' find orthogonal eigenvectors to +ve 'm' eigenvectors
        if m < 0:
            eigenvectors *= 1j

        return eigenvalues, eigenvectors

    @pydantic.field_validator("order")
    def _check_order(
        cls,  # noqa: ANN101
        v: int | npt.NDArray[np.int_] | None,
        info: pydantic.ValidationInfo,
    ) -> int | npt.NDArray[np.int_] | None:
        if v is not None and (np.abs(v) >= info.data["L"]).any():
            msg = f"Order magnitude should be less than {info.data['L']}"
            raise ValueError(msg)
        return v

    @pydantic.field_validator("theta_max")
    def _check_theta_max(
        cls,  # noqa: ANN101
        v: float,
    ) -> float:
        if v == 0:
            msg = "theta_max cannot be zero"
            raise ValueError(msg)
        return v



"""Contains the `HarmonicGaussian` class."""
import numpy as np
import numpy.typing as npt
import pydantic
import typing_extensions

import pyssht as ssht

import sleplet._string_methods
import sleplet._validation
from sleplet.functions.flm import Flm


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class HarmonicGaussian(Flm):
    r"""
    Create a harmonic Gaussian
    \(\exp(-(\frac{{\ell}^{2}}{2\sigma_{\ell}^{2}}
    + \frac{{m}^{2}}{2\sigma_{m}^{2}}))\).
    """

    l_sigma: float = 10
    r"""Sets the \(\sigma_{\ell}\) value."""
    m_sigma: float = 10
    r"""Sets the \(\sigma_{m}\) value."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        flm = np.zeros(self.L**2, dtype=np.complex_)
        for ell in range(self.L):
            upsilon_l = np.exp(-((ell / self.l_sigma) ** 2) / 2)
            for m in range(-ell, ell + 1):
                ind = ssht.elm2ind(ell, m)
                flm[ind] = upsilon_l * np.exp(-((m / self.m_sigma) ** 2) / 2)
        return flm

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{sleplet._string_methods.filename_args(self.l_sigma, 'lsig')}"
            f"{sleplet._string_methods.filename_args(self.m_sigma, 'msig')}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return False

    def _set_spin(self: typing_extensions.Self) -> int:
        return 0

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 2
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.l_sigma, self.m_sigma = (
                np.float_power(10, x) for x in self.extra_args
            )



"""Methods to handle noise in Fourier or wavelet space."""
import logging

import numpy as np
import numpy.typing as npt

import pyssht as ssht

import sleplet._vars
import sleplet.harmonic_methods
import sleplet.meshes.mesh_slepian
import sleplet.slepian_methods
from sleplet.slepian.slepian_functions import SlepianFunctions

_logger = logging.getLogger(__name__)


def _signal_power(signal: npt.NDArray[np.complex_ | np.float_]) -> float:
    """Compute the power of the signal."""
    return (np.abs(signal) ** 2).sum()


def compute_snr(
    signal: npt.NDArray[np.complex_ | np.float_],
    noise: npt.NDArray[np.complex_ | np.float_],
    signal_type: str,
) -> float:
    """
    Compute the signal to noise ratio.

    Args:
        signal: The unnoised signal.
        noise: The noised signal.
        signal_type: Specifier to improve logging.

    Returns:
        The signal-to-noise value of the noised signal.
    """
    snr = 10 * np.log10(_signal_power(signal) / _signal_power(noise))
    msg = f"{signal_type} SNR: {snr:.2f}"
    _logger.info(msg)
    return snr


def compute_sigma_noise(
    signal: npt.NDArray[np.complex_ | np.float_],
    snr_in: float,
    *,
    denominator: int | None = None,
) -> float:
    """
    Compute the standard deviation of the noise.

    Args:
        signal: The noised signal.
        snr_in: The parameter controlling the signal-to-noise.
        denominator: Coefficients to use in computing the noise, defaults to all.

    Returns:
        The standard deviation of the noise.
    """
    if denominator is None:
        denominator = signal.shape[0]
    return np.sqrt(10 ** (-snr_in / 10) * _signal_power(signal) / denominator)


def _create_noise(
    L: int,
    signal: npt.NDArray[np.complex_ | np.float_],
    snr_in: float,
) -> npt.NDArray[np.complex_]:
    """Compute Gaussian white noise."""
    # set random seed
    rng = np.random.default_rng(sleplet._vars.RANDOM_SEED)

    # initialise
    nlm = np.zeros(L**2, dtype=np.complex_)

    # std dev of the noise
    sigma_noise = compute_sigma_noise(signal, snr_in)

    # compute noise
    for ell in range(L):
        ind = ssht.elm2ind(ell, 0)
        nlm[ind] = sigma_noise * rng.standard_normal()
        for m in range(1, ell + 1):
            ind_pm = ssht.elm2ind(ell, m)
            ind_nm = ssht.elm2ind(ell, -m)
            nlm[ind_pm] = (
                sigma_noise
                / np.sqrt(2)
                * (rng.standard_normal() + 1j * rng.standard_normal())
            )
            nlm[ind_nm] = (-1) ** m * nlm[ind_pm].conj()
    return nlm


def _create_slepian_noise(
    L: int,
    slepian_signal: npt.NDArray[np.complex_ | np.float_],
    slepian: SlepianFunctions,
    snr_in: float,
) -> npt.NDArray[np.complex_]:
    """Compute Gaussian white noise in Slepian space."""
    flm = ssht.forward(
        sleplet.slepian_methods.slepian_inverse(slepian_signal, L, slepian),
        L,
        Method=sleplet._vars.SAMPLING_SCHEME,
    )
    nlm = _create_noise(L, flm, snr_in)
    return sleplet.slepian_methods.slepian_forward(L, slepian, flm=nlm)


def _perform_hard_thresholding(
    f: npt.NDArray[np.complex_ | np.float_],
    sigma_j: float | npt.NDArray[np.float_],
    n_sigma: int,
) -> npt.NDArray[np.complex_]:
    """Set pixels in pixel space to zero if the magnitude is less than the threshold."""
    threshold = n_sigma * sigma_j
    return np.where(np.abs(f) < threshold, 0, f)


def harmonic_hard_thresholding(
    L: int,
    wav_coeffs: npt.NDArray[np.complex_],
    sigma_j: npt.NDArray[np.float_],
    n_sigma: int,
) -> npt.NDArray[np.complex_]:
    r"""
    Perform thresholding in harmonic space.

    Args:
        L: The spherical harmonic bandlimit.
        wav_coeffs: The harmonic wavelet coefficients.
        sigma_j: The wavelet standard deviation \(\sigma_{j}\).
        n_sigma: The number of \(\sigma_{j}\) to threshold.

    Returns:
        The thresholded wavelet coefficients.
    """
    _logger.info("begin harmonic hard thresholding")
    for j, coefficient in enumerate(wav_coeffs[1:]):
        msg = f"start Psi^{j + 1}/{len(wav_coeffs)-1}"
        _logger.info(msg)
        f = ssht.inverse(coefficient, L, Method=sleplet._vars.SAMPLING_SCHEME)
        f_thresholded = _perform_hard_thresholding(f, sigma_j[j], n_sigma)
        wav_coeffs[j + 1] = ssht.forward(
            f_thresholded,
            L,
            Method=sleplet._vars.SAMPLING_SCHEME,
        )
    return wav_coeffs


def slepian_wavelet_hard_thresholding(
    L: int,
    wav_coeffs: npt.NDArray[np.complex_ | np.float_],
    sigma_j: npt.NDArray[np.float_],
    n_sigma: int,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.complex_ | np.float_]:
    r"""
    Perform thresholding in Slepian wavelet space.

    Args:
        L: The spherical harmonic bandlimit.
        wav_coeffs: The Slepian wavelet coefficients
        sigma_j: The wavelet standard deviation \(\sigma_{j}\).
        n_sigma: The number of \(\sigma_{j}\) to threshold.
        slepian: The given Slepian object.

    Returns:
        The hard thresholded Slepian wavelet coefficients.
    """
    _logger.info("begin Slepian hard thresholding")
    for j, coefficient in enumerate(wav_coeffs):
        msg = f"start Psi^{j + 1}/{len(wav_coeffs)}"
        _logger.info(msg)
        f = sleplet.slepian_methods.slepian_inverse(coefficient, L, slepian)
        f_thresholded = _perform_hard_thresholding(f, sigma_j[j], n_sigma)
        wav_coeffs[j] = sleplet.slepian_methods.slepian_forward(
            L,
            slepian,
            f=f_thresholded,
        )
    return wav_coeffs


def slepian_function_hard_thresholding(
    L: int,
    coefficients: npt.NDArray[np.complex_ | np.float_],
    sigma: float,
    n_sigma: int,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.complex_]:
    r"""
    Perform thresholding in Slepian space.

    Args:
        L: The spherical harmonic bandlimit.
        coefficients: The Slepian coefficients.
        sigma: The standard deviation of the noise \(\sigma\).
        n_sigma: The number of \(\sigma\) to threshold.
        slepian: The given Slepian object.

    Returns:
        The thresholded Slepian coefficients.
    """
    _logger.info("begin Slepian hard thresholding")
    f = sleplet.slepian_methods.slepian_inverse(coefficients, L, slepian)
    f_thresholded = _perform_hard_thresholding(f, sigma, n_sigma)
    return sleplet.slepian_methods.slepian_forward(L, slepian, f=f_thresholded)


def _compute_sigma_j(
    signal: npt.NDArray[np.complex_ | np.float_],
    psi_j: npt.NDArray[np.complex_],
    snr_in: float,
) -> npt.NDArray[np.float_]:
    """Compute sigma_j for wavelets used in denoising the signal."""
    sigma_noise = compute_sigma_noise(signal, snr_in)
    wavelet_power = (np.abs(psi_j) ** 2).sum(axis=1)
    return sigma_noise * np.sqrt(wavelet_power)


def _compute_slepian_sigma_j(
    L: int,
    signal: npt.NDArray[np.complex_ | np.float_],
    psi_j: npt.NDArray[np.float_],
    snr_in: float,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.float_]:
    """Compute sigma_j for wavelets used in denoising the signal."""
    sigma_noise = compute_sigma_noise(signal, snr_in, denominator=L**2)
    s_p = sleplet.slepian_methods.compute_s_p_omega(L, slepian)
    psi_j_reshape = psi_j[:, : slepian.N, np.newaxis, np.newaxis]
    wavelet_power = (np.abs(psi_j_reshape) ** 2 * np.abs(s_p) ** 2).sum(axis=1)
    return sigma_noise * np.sqrt(wavelet_power)


def _create_mesh_noise(
    u_i: npt.NDArray[np.complex_ | np.float_],
    snr_in: float,
) -> npt.NDArray[np.float_]:
    """Compute Gaussian white noise."""
    # set random seed
    rng = np.random.default_rng(sleplet._vars.RANDOM_SEED)

    # initialise
    n_i = np.zeros(u_i.shape[0])

    # std dev of the noise
    sigma_noise = compute_sigma_noise(u_i, snr_in)

    # compute noise
    for i in range(u_i.shape[0]):
        n_i[i] = sigma_noise * rng.standard_normal()
    return n_i


def _create_slepian_mesh_noise(
    mesh_slepian: "sleplet.meshes.mesh_slepian.MeshSlepian",
    slepian_signal: npt.NDArray[np.complex_ | np.float_],
    snr_in: float,
) -> npt.NDArray[np.float_]:
    """Compute Gaussian white noise in Slepian space."""
    u_i = sleplet.harmonic_methods.mesh_forward(
        mesh_slepian.mesh,
        sleplet.slepian_methods.slepian_mesh_inverse(
            mesh_slepian,
            slepian_signal,
        ),
    )
    n_i = _create_mesh_noise(u_i, snr_in)
    return sleplet.slepian_methods.slepian_mesh_forward(
        mesh_slepian,
        u_i=n_i,
    )


def compute_slepian_mesh_sigma_j(
    mesh_slepian: "sleplet.meshes.mesh_slepian.MeshSlepian",
    signal: npt.NDArray[np.complex_ | np.float_],
    psi_j: npt.NDArray[np.float_],
    snr_in: float,
) -> npt.NDArray[np.float_]:
    r"""
    Compute \(\sigma_{j}\) for wavelets used in denoising the signal.

    Args:
        mesh_slepian: The Slepian mesh object containing the eigensolutions.
        signal: The noised signal.
        psi_j: The Slepian wavelet coefficients.
        snr_in: The parameter controlling the signal-to-noise ratio.

    Returns:
        The standard deviation of the noise.
    """
    sigma_noise = compute_sigma_noise(
        signal,
        snr_in,
        denominator=mesh_slepian.slepian_eigenvalues.shape[0],
    )
    s_p = sleplet.slepian_methods._compute_mesh_s_p_pixel(mesh_slepian)
    psi_j_reshape = psi_j[:, : mesh_slepian.N, np.newaxis]
    wavelet_power = (np.abs(psi_j_reshape) ** 2 * np.abs(s_p) ** 2).sum(axis=1)
    return sigma_noise * np.sqrt(wavelet_power)


def slepian_mesh_hard_thresholding(
    mesh_slepian: "sleplet.meshes.mesh_slepian.MeshSlepian",
    wav_coeffs: npt.NDArray[np.complex_ | np.float_],
    sigma_j: npt.NDArray[np.float_],
    n_sigma: int,
) -> npt.NDArray[np.complex_ | np.float_]:
    r"""
    Perform thresholding in Slepian space of the mesh.

    Args:
        mesh_slepian: The Slepian mesh object containing the eigensolutions.
        wav_coeffs: The Slepian wavelet coefficients of the mesh.
        sigma_j: The wavelet standard deviation \(\sigma_{j}\).
        n_sigma: The number of \(\sigma\) to threshold.

    Returns:
        The thresholded wavelet coefficients of the mesh.
    """
    _logger.info("begin Slepian mesh hard thresholding")
    for j, coefficient in enumerate(wav_coeffs):
        msg = f"start Psi^{j + 1}/{len(wav_coeffs)}"
        _logger.info(msg)
        f = sleplet.slepian_methods.slepian_mesh_inverse(mesh_slepian, coefficient)
        f_thresholded = _perform_hard_thresholding(f, sigma_j[j], n_sigma)
        wav_coeffs[j] = sleplet.slepian_methods.slepian_mesh_forward(
            mesh_slepian,
            u=f_thresholded,
        )
    return wav_coeffs



"""Contains the `Ridgelets` class."""
import logging

import numpy as np
import numpy.typing as npt
import pydantic
import scipy.special
import typing_extensions

import pys2let
import pyssht as ssht

import sleplet._string_methods
import sleplet._validation
import sleplet.wavelet_methods
from sleplet.functions.flm import Flm

_logger = logging.getLogger(__name__)


@pydantic.dataclasses.dataclass(config=sleplet._validation.validation, kw_only=True)
class Ridgelets(Flm):
    """
    Crates scale-discretised wavelets. As seen in
    <https://arxiv.org/abs/1510.01595>.
    """

    B: int = 3
    r"""The wavelet parameter. Represented as \(\lambda\) in the papers."""
    j_min: int = 2
    r"""The minimum wavelet scale. Represented as \(J_{0}\) in the papers."""
    j: int | None = None
    """Option to select a given wavelet. `None` indicates the scaling function,
    whereas would correspond to the selected `j_min`."""
    spin: int = 2
    """Spin value."""

    def __post_init__(self: typing_extensions.Self) -> None:
        super().__post_init__()

    def _create_coefficients(
        self: typing_extensions.Self,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        _logger.info("start computing wavelets")
        self.wavelets = self._create_wavelets()
        _logger.info("finish computing wavelets")
        jth = 0 if self.j is None else self.j + 1
        return self.wavelets[jth]

    def _create_name(self: typing_extensions.Self) -> str:
        return (
            f"{sleplet._string_methods._convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{sleplet._string_methods.filename_args(self.B, 'B')}"
            f"{sleplet._string_methods.filename_args(self.j_min, 'jmin')}"
            f"{sleplet._string_methods.filename_args(self.spin, 'spin')}"
            f"{sleplet._string_methods.wavelet_ending(self.j_min, self.j)}"
        )

    def _set_reality(self: typing_extensions.Self) -> bool:
        return False

    def _set_spin(self: typing_extensions.Self) -> int:
        return self.spin

    def _setup_args(self: typing_extensions.Self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 4
            if len(self.extra_args) != num_args:
                msg = f"The number of extra arguments should be {num_args}"
                raise ValueError(msg)
            self.B, self.j_min, self.spin, self.j = self.extra_args

    def _create_wavelets(self: typing_extensions.Self) -> npt.NDArray[np.complex_]:
        """Compute all wavelets."""
        ring_lm = self._compute_ring()
        kappas = sleplet.wavelet_methods.create_kappas(self.L, self.B, self.j_min)
        wavelets = np.zeros((kappas.shape[0], self.L**2), dtype=np.complex_)
        for ell in range(self.L):
            ind = ssht.elm2ind(ell, 0)
            wavelets[0, ind] = kappas[0, ell] * ring_lm[ind]
            wavelets[1:, ind] = kappas[1:, ell] * ring_lm[ind] / np.sqrt(2 * np.pi)
        return wavelets

    def _compute_ring(self: typing_extensions.Self) -> npt.NDArray[np.complex_]:
        """Compute ring in harmonic space."""
        ring_lm = np.zeros(self.L**2, dtype=np.complex_)
        for ell in range(abs(self.spin), self.L):
            logp2 = (
                scipy.special.gammaln(ell + self.spin + 1)
                - ell * np.log(2)
                - scipy.special.gammaln((ell + self.spin) / 2 + 1)
                - scipy.special.gammaln((ell - self.spin) / 2 + 1)
            )
            p0 = np.real((-1) ** ((ell + self.spin) / 2)) * np.exp(logp2)
            ind = ssht.elm2ind(ell, 0)
            ring_lm[ind] = (
                2
                * np.pi
                * np.sqrt((2 * ell + 1) / (4 * np.pi))
                * p0
                * (-1) ** self.spin
                * np.sqrt(
                    np.exp(
                        scipy.special.gammaln(ell - self.spin + 1)
                        - scipy.special.gammaln(ell + self.spin + 1),
                    ),
                )
            )
        return ring_lm

    @pydantic.field_validator("j")
    def _check_j(
        cls,  # noqa: ANN101
        v: int | None,
        info: pydantic.ValidationInfo,
    ) -> int | None:
        j_max = pys2let.pys2let_j_max(
            info.data["B"],
            info.data["L"],
            info.data["j_min"],
        )
        if v is not None and v < 0:
            msg = "j should be positive"
            raise ValueError(msg)
        if v is not None and v > j_max - info.data["j_min"]:
            msg = (
                "j should be less than j_max - j_min: "
                f"{j_max - info.data['j_min'] + 1}"
            )
            raise ValueError(msg)
        return v
